# This is a Shiny web application. You can run the application in RStudio
# by clicking the 'Run App' button at the top right of the editor pane.

library(shiny)
library(shinyjs)
library(scatterD3)
library(htmlwidgets) # for saveWidget()
library(magrittr)

vars <- load("data.rda")
## ByType4.df, ByType32.df, dim2label, symbols.vec, 
## rainbow32.vec, rainbow20.vec, rainbow12.vec, 
## types.variety, types.mode, 
## types.short32, types.short20, types.short12, 
## types.textcat32, types.textcat20, types.textcat12

## mappings between codes/short names and long names
mk.map <- function (x, y) structure(y, names=x)
inverse.map <- function (x) mk.map(x, names(x))

# dim2label mapping has been loaded from "data.rda"
label2dim <- inverse.map(dim2label)
label2dim2 <- inverse.map(dim2label2) # shorter labels

label2cat.32 <- mk.map(types.short32, types.textcat32)
cat2label.32 <- inverse.map(label2cat.32)

label2cat.20 <- mk.map(types.short20, types.textcat20)
cat2label.20 <- inverse.map(label2cat.20)

label2cat.12 <- mk.map(types.short12, types.textcat12)
cat2label.12 <- inverse.map(label2cat.12)

zoom.set.code <- "
Shiny.setInputValue('xmin', %g);
Shiny.setInputValue('xmax', %g);
Shiny.setInputValue('ymin', %g);
Shiny.setInputValue('ymax', %g);
"
zoom.default <- list(xmin=-3.1, xmax=3.1, ymin=-3.1, ymax=3.1)

## define presets (which set all listed inputs to the specified values)
source("presets.R", local=TRUE) # load user presets from external file
source("figures.R", local=TRUE) # presets for figures in the paper
Builtin <- list(
    default = list(name="default settings",
                   lda="lda20", y="LD1", x="LD2", granularity="n20", ellipses=FALSE, focus_mode="both", focus_variety="all",
                   show_textcat12=label2cat.12, show_textcat20=label2cat.20, show_textcat32=label2cat.32, 
                   pointsize=4, transitions=FALSE, viewport=zoom.default),
    allcat = list(name="show all categories",
                  show_textcat12=label2cat.12, show_textcat20=label2cat.20, show_textcat32=label2cat.32),
    nocat = list(name="hide all categories",
                 show_textcat12=character(), show_textcat20=character(), show_textcat32=character())
)
Presets <- c(Builtin, Figures, Presets)
presets.choices <- mk.map(sapply(Presets, function (x) x$name), names(Presets))

##
## UI
##
ui <- fluidPage(
    useShinyjs(),
    tags$script(JS("$(document).on('shiny:sessioninitialized', function (event) {$('#reset_zoom').click();});")), # so plot dimensions are 
    titlePanel("Neumann & Evert (2021): Scatterplots of latent dimensions"),
    # div(class="row",
    #     div(class="col-md-12",
    #         div(class="alert alert-warning alert-dismissible",
    #             HTML('<button type="button" class="close" data-dismiss="alert" aria-label="Close"><span aria-hidden="true">&times;</span></button>'),
    #             HTML("<strong>Alert box</strong>")))),
    sidebarLayout(
        sidebarPanel(
            radioButtons("lda", NULL,
                        choices=c("20-cat LDA"="lda20", "32-cat LDA"="lda32"),
                        selected="lda20", inline=TRUE),
            selectInput("y", "y-axis (vertical)",
                        choices = label2dim2,
                        selected = "LD1"),
            selectInput("x", "x-axis (horizontal)",
                        choices = label2dim2,
                        selected = "LD2"),
            radioButtons("granularity", "number of text categories",
                         choices=c("12"="n12", "20"="n20", "32"="n32"), selected="n20", inline=TRUE),
            checkboxInput("ellipses", "confidence ellipses", value = TRUE),
            tags$hr(),
            radioButtons("focus_mode", "text mode",
                         choices=c("both"="both", "written"="W", "spoken"="S"), selected="both",
                         inline=TRUE),
            radioButtons("focus_variety", "language variety",
                         choices=c("all", names(symbols.vec)), selected="all",
                         inline=TRUE),
            conditionalPanel(condition="input.granularity == 'n12'",
                             checkboxGroupInput("show_textcat12", "text categories",
                                                choices=label2cat.12, selected=label2cat.12,
                                                inline=TRUE)),
            conditionalPanel(condition="input.granularity == 'n20'",
                             checkboxGroupInput("show_textcat20", "text categories",
                                                choices=label2cat.20, selected=label2cat.20,
                                                inline=TRUE)),
            conditionalPanel(condition="input.granularity == 'n32'",
                             checkboxGroupInput("show_textcat32", "text categories",
                                                choices=label2cat.32, selected=label2cat.32,
                                                inline=TRUE)),
            tags$hr(style="height: 3px; background: #DDD;"),
            sliderInput("pointsize", "point size",
                        min = 1, max = 6, value = 4, step = .1),
            checkboxInput("transitions", "use transitions", value = FALSE),
            tags$p(#actionButton("scatterD3-reset-zoom", HTML("<span class='glyphicon glyphicon-search' aria-hidden='true'></span> Reset Zoom")),
                   actionButton("reset_zoom", HTML("<span class='glyphicon glyphicon-search' aria-hidden='true'></span> Reset Zoom"),
                          onclick = paste(
                              sprintf(zoom.set.code, zoom.default$xmin, zoom.default$xmax, zoom.default$ymin, zoom.default$ymax),
                              "return true;")),
                   tags$a(id = "scatterD3-svg-export", href = "#", class = "btn btn-default", 
                          HTML("<span class='glyphicon glyphicon-save' aria-hidden='true'></span> Download SVG")),
                   # actionButton("scatterD3-lasso-toggle", HTML("<span class='glyphicon glyphicon-screenshot' aria-hidden='true'></span> Toggle Lasso"),
                   # "data-toggle" = "button")),
                   downloadButton("download_plot", "Download Interactive Plot", icon("download"))
            ),
            tags$hr(style="height: 3px; background: #DDD;"),
            selectInput("preset", "choose a preset:", 
                        choices=presets.choices, selected="default"),
            tags$hr(style="height: 3px; background: #DDD;"),
            downloadButton("save_preset", "Save Preset", icon("download")),
            checkboxGroupInput("save_preset_options", "", 
                               choices=c("include all textcat selections"="allcat", 
                                         "include current zoom"="viewport"),
                               selected=NULL, inline=FALSE),
            width = 3
        ),
        mainPanel(
            scatterD3Output("scatterPlot", width = "100%", height="1200px"),
            width = 9
        )
    )
)

##
## Server
##
server <- function(input, output, session) {
    updateInput <- function (name, val, type=c("radio", "select", "checkbox", "group", "slider")) {
        type <- match.arg(type)
        if (is.null(val)) return()
        switch(type,
               radio = updateRadioButtons(session, name, selected=val),
               select = updateSelectInput(session, name, selected=val),
               checkbox = updateCheckboxInput(session, name, value=val),
               group = updateCheckboxGroupInput(session, name, selected=val),
               slider = updateSliderInput(session, name, value=val))
    }
    
    observe({
        preset <- Presets[[input$preset]]
        for (par in c("x", "y")) updateInput(par, preset[[par]], "select")
        for (par in c("lda", "granularity", "focus_mode", "focus_variety")) updateInput(par, preset[[par]], "radio")
        for (par in c("ellipses", "transitions")) updateInput(par, preset[[par]], "checkbox")
        for (par in c("show_textcat12", "show_textcat20", "show_textcat32")) updateInput(par, preset[[par]], "group")
        for (par in c("pointsize")) updateInput(par, preset[[par]], "slider")
        vp <- preset$viewport
        if (!is.null(vp)) runjs(sprintf(zoom.set.code, vp$xmin, vp$xmax, vp$ymin, vp$ymax))
    })
    
    output$save_preset <- downloadHandler(
        filename = "scatterplot_preset.R",
        content = function (filename) {
            preset <- list(
                name = "",
                x = isolate(input$x),
                y = isolate(input$y),
                lda = isolate(input$lda),
                granularity = isolate(input$granularity),
                focus_mode = isolate(input$focus_mode),
                focus_variety = isolate(input$focus_variety),
                ellipses = isolate(input$ellipses),
                transitions = isolate(input$transitions),
                pointsize = isolate(input$pointsize)
            )
            opts <- isolate(input$save_preset_options)
            all.cat <- ("allcat" %in% opts) # whether to save all category selections
            if (all.cat || preset$granularity == "n12") preset$show_textcat12 <- isolate(input$show_textcat12)
            if (all.cat || preset$granularity == "n20") preset$show_textcat20 <- isolate(input$show_textcat20)
            if (all.cat || preset$granularity == "n32") preset$show_textcat32 <- isolate(input$show_textcat32)
            if ("viewport" %in% opts) preset$viewport <- list(
                xmin=isolate(input$xmin), xmax=isolate(input$xmax),
                ymin=isolate(input$ymin), ymax=isolate(input$ymax))
            dump("preset", file=filename)
        },
        contentType = "text/plain"
    )
    
    data <- reactive({
        data.df <- if (input$lda == "lda32") ByType32.df else ByType4.df
        if (input$focus_mode != "both") data.df %<>% subset(mode == input$focus_mode)
        data.df <- switch(
            input$granularity,
            n12 = subset(data.df, textcat12 %in% input$show_textcat12),
            n20 = subset(data.df, textcat20 %in% input$show_textcat20),
            n32 = subset(data.df, textcat32 %in% input$show_textcat32))
        data.df %<>% droplevels()
        if (input$focus_variety != "all") {
            data.df %<>% subset(variety == input$focus_variety) # no droplevels so we don't lose the legend
        }
        data.df
    })
    
    lasso.callback <- "
function(sel) {
    /* this is just a dummy implementation to be replaced later */
    pts = sel.data().map(function(d) {return d.key_var}); 
    alert(pts.length + ' texts selected');
}"
    zoom.callback <- "
function(xmin, xmax, ymin, ymax) {
    Shiny.setInputValue('xmin', xmin);
    Shiny.setInputValue('xmax', xmax);
    Shiny.setInputValue('ymin', ymin);
    Shiny.setInputValue('ymax', ymax);
}"
    
    ## create the D3 scatterplot object, so it can either be displayed or saved to a file
    D3plot <- reactive({
        input$reset_zoom # so redisplay is triggered after resetting zoom
        col.vec <- switch(
            input$granularity,
            n12 = data()$textcat12,
            n20 = data()$textcat20,
            n32 = data()$textcat32)
        col.values <- switch(
            input$granularity,
            n12 = rainbow12.vec[levels(data()$textcat12)],
            n20 = rainbow20.vec[levels(data()$textcat20)],
            n32 = rainbow32.vec[levels(data()$textcat32)])
        scatterD3(y = data()[, input$y],
                  x = data()[, input$x],
                  fixed = TRUE,
                  xlab = dim2label[[input$x]],
                  ylab = dim2label[[input$y]],
                  xlim = c(isolate(input$xmin), isolate(input$xmax)),
                  ylim = c(isolate(input$ymin), isolate(input$ymax)),
                  col_var = col.vec,
                  col_lab = "Text Category",
                  colors = col.values,
                  symbol_var = data()$variety,
                  symbol_lab = "Variety",
                  symbols = symbols.vec,
                  point_size = 2 ^ input$pointsize,
                  tooltips = TRUE,
                  tooltip_text = data()$tooltips,
                  key_var = rownames(data()),
                  ellipses = input$ellipses,
                  transitions = input$transitions,
                  lasso = TRUE,
                  lasso_callback = lasso.callback,
                  zoom_callback = zoom.callback,
                  axes_font_size = "160%",
                  legend_font_size = "140%",
                  legend_width = 200,
                  left_margin = 40,
                  menu = FALSE)
    })
    
    output$scatterPlot <- renderScatterD3({
        D3plot()
    })
    
    output$download_plot <- downloadHandler(
        filename = function () {
            cat.list <- switch(
                input$granularity,
                n12 = if (setequal(input$show_textcat12, label2cat.12)) "all" else cat2label.12[input$show_textcat12],
                n20 = if (setequal(input$show_textcat20, label2cat.20)) "all" else cat2label.20[input$show_textcat20],
                n32 = if (setequal(input$show_textcat32, label2cat.32)) "all" else cat2label.32[input$show_textcat32])
            cat.list <- if (length(cat.list) > 0) paste(cat.list, collapse="-") else "none"
            sprintf("ICE-scatter_%s_%s_%s_%s-%s%s_cat%d=%s.html",
                    input$lda,
                    input$focus_variety,
                    input$focus_mode,
                    input$y, input$x,
                    if (input$ellipses) "_conf" else "",
                    switch(input$granularity, n12=12, n20=20, n32=32),
                    cat.list)
        },
        content = function (filename) {
            mode.txt <- switch(input$focus_mode, both="both", W="written", S="spoken")
            title.txt <- sprintf("ICE-%s – %s vs. %s – %s",
                                 input$focus_variety,
                                 dim2label[input$y], dim2label[input$x],
                                 mode.txt)
            saveWidget(D3plot(), file=filename, title=title.txt, selfcontained=TRUE)
        },
        contentType = "application/octet-stream"
    )
}

# Run the application 
shinyApp(ui = ui, server = server)
