# This is a Shiny web application. You can run the application in RStudio
# by clicking the 'Run App' button at the top right of the editor pane.

library(shiny)
library(shinyjs)
library(ggplot2)
library(reshape2)
library(magrittr)
source("utilities.R")

vars <- load("data.rda")
# ZL, Meta, Dimensions, Dimensions32, idx.weights, feature.names, 
# rainbow32.vec, rainbow20.vec, rainbow12.vec, dim2label, 
# types.short32, types.short20, types.short12, 
# types.textcat32, types.textcat20, types.textcat12

## mappings between codes/short names and long names
mk.map <- function (x, y) structure(y, names=x)
inverse.map <- function (x) mk.map(x, names(x))

# dim2label mapping has been loaded from "data.rda"
label2dim <- inverse.map(dim2label)
label2dim2 <- inverse.map(dim2label2) # shorter labels

label2cat.32 <- mk.map(types.short32, types.textcat32)
cat2label.32 <- inverse.map(label2cat.32)

label2cat.20 <- mk.map(types.short20, types.textcat20)
cat2label.20 <- inverse.map(label2cat.20)

label2cat.12 <- mk.map(types.short12, types.textcat12)
cat2label.12 <- inverse.map(label2cat.12)

## persistent cache for generated plots
persistent.cache <- TRUE  # set to FALSE for development
max.files <- 1000         # how many plots can be cached (ca. 0.2 MB / plot)
if (persistent.cache) shinyOptions(cache = diskCache("./cache", max_n = max.files)) # cache up to 1000 plots = ca. 200 MB


## define presets (which set all listed inputs to the specified values)
source("presets.R", local=TRUE) # load user presets from external file

mode2cats <- function (granularity, mode) switch(
  granularity,
  n12 = unique(Meta$textcat12[Meta$mode == mode]),
  n20 = unique(Meta$textcat20[Meta$mode == mode]),
  n32 = unique(Meta$textcat32[Meta$mode == mode]),
  stop("invalid granularity=", granularity))
Builtin <- list(
  default = list(name="default settings",
                 lda="lda20", dim="LD1", y="contribution", granularity="n20", focus_variety="all",
                 show_textcat12=character(), show_textcat20=character(), show_textcat32=character()),
  allcat = list(name="include all categories",
                show_textcat12=label2cat.12, show_textcat20=label2cat.20, show_textcat32=label2cat.32),
  allwritten = list(name="include all written categories",
                    show_textcat12=mode2cats("n12", "W"), show_textcat20=mode2cats("n20", "W"), show_textcat32=mode2cats("n32", "W")),
  allspoken = list(name="include all spoken categories",
                   show_textcat12=mode2cats("n12", "S"), show_textcat20=mode2cats("n20", "S"), show_textcat32=mode2cats("n32", "S")),
  nocat = list(name="hide all categories",
               show_textcat12=character(), show_textcat20=character(), show_textcat32=character())
)

Presets <- c(Builtin, Presets)
presets.choices <- mk.map(sapply(Presets, function (x) x$name), names(Presets))


##
## UI
##
ui <- fluidPage(
  useShinyjs(),
  titlePanel("Neumann & Evert (2021): Feature weights and contributions"),
  # div(class="row",
  #     div(class="col-md-12",
  #         div(class="alert alert-warning alert-dismissible",
  #             HTML('<button type="button" class="close" data-dismiss="alert" aria-label="Close"><span aria-hidden="true">&times;</span></button>'),
  #             HTML("<strong>Alert box</strong>")))),
  sidebarLayout(
    sidebarPanel(
      radioButtons("lda", NULL,
                   choices=c("20-cat LDA"="lda20", "32-cat LDA"="lda32"),
                   selected="lda20", inline=TRUE),
      selectInput("dim", "dimension",
                  choices = label2dim2,
                  selected = "LD1"),
      selectInput("y", "y-axis",
                  choices = c("features", "weighted", "contribution"),
                  selected = "contribution"),
      radioButtons("granularity", "number of text categories",
                   choices=c("12"="n12", "20"="n20", "32"="n32"), selected="n20", inline=TRUE),
      conditionalPanel(condition="input.granularity == 'n12'",
                       checkboxGroupInput("show_textcat12", "text categories",
                                          choices=label2cat.12, selected=NULL,
                                          inline=TRUE)),
      conditionalPanel(condition="input.granularity == 'n20'",
                       checkboxGroupInput("show_textcat20", "text categories",
                                          choices=label2cat.20, selected=NULL,
                                          inline=TRUE)),
      conditionalPanel(condition="input.granularity == 'n32'",
                       checkboxGroupInput("show_textcat32", "text categories",
                                          choices=label2cat.32, selected=NULL,
                                          inline=TRUE)),
      radioButtons("focus_variety", "language variety",
                   choices=c("all", levels(Meta$variety)), selected="all",
                   inline=TRUE),
      tags$hr(style="height: 3px; background: #DDD;"),
      radioButtons("plot_size", "display size",
                   choices=c("S", "M", "L", "XL"), selected="M",
                   inline=TRUE),
      checkboxInput("use_legend", "legend in sidebar", value=TRUE),
      tags$hr(style="height: 3px; background: #DDD;"),
      selectInput("preset", "choose a preset:", 
                  choices=presets.choices, selected="default"),
      downloadButton("downloadPDF", "Download PDF", icon("download")),
      tags$hr(style="height: 3px; background: #DDD;"),
      downloadButton("save_preset", "Save Preset", icon("download")),
      checkboxGroupInput("save_preset_options", "", 
                         choices=c("include all textcat selections"="allcat"),
                         selected=NULL, inline=FALSE),
      width = 3
    ),
    mainPanel(
      plotOutput("boxplot", width = "1050px", height="700px"),
      plotOutput("densities", width = "1050px", height="200px"),
      width = 9
    )
  )
)


##
## Server
##
server <- function(input, output, session) {
  updateInput <- function (name, val, type=c("radio", "select", "checkbox", "group", "slider")) {
    type <- match.arg(type)
    if (is.null(val)) return()
    switch(type,
           radio = updateRadioButtons(session, name, selected=val),
           select = updateSelectInput(session, name, selected=val),
           checkbox = updateCheckboxInput(session, name, value=val),
           group = updateCheckboxGroupInput(session, name, selected=val),
           slider = updateSliderInput(session, name, value=val))
  }
  
  observe({
    preset <- Presets[[input$preset]]
    for (par in c("dim", "y")) updateInput(par, preset[[par]], "select")
    for (par in c("lda", "granularity", "focus_variety")) updateInput(par, preset[[par]], "radio")
    for (par in c("show_textcat12", "show_textcat20", "show_textcat32")) updateInput(par, preset[[par]], "group")
  })

  output$save_preset <- downloadHandler(
    filename = "boxplot_preset.R",
    content = function (filename) {
      preset <- list(
        name = "",
        lda = isolate(input$lda),
        dim = isolate(input$dim),
        y = isolate(input$y),
        granularity = isolate(input$granularity),
        focus_variety = isolate(input$focus_variety)
      )
      opts <- isolate(input$save_preset_options)
      all.cat <- ("allcat" %in% opts) # whether to save all category selections
      coalesce <- function (x, default=character()) if (is.null(x)) default else x
      if (all.cat || preset$granularity == "n12") preset$show_textcat12 <- coalesce(isolate(input$show_textcat12))
      if (all.cat || preset$granularity == "n20") preset$show_textcat20 <- coalesce(isolate(input$show_textcat20))
      if (all.cat || preset$granularity == "n32") preset$show_textcat32 <- coalesce(isolate(input$show_textcat32))
      dump("preset", file=filename)
    },
    contentType = "text/plain"
  )
    
  Data <- reactive({
    Dims <- switch(input$lda,
                   lda20 = Dimensions,
                   lda32 = Dimensions32)
    if (input$focus_variety != "all") {
      idx <- Meta$variety == input$focus_variety
      list(M=ZL[idx, ], Meta=Meta[idx, ], Dims=Dims)
    }
    else {
      list(M=ZL[, ], Meta=Meta, Dims=Dims)
    }
  })
  
  plot.width <-    c(S=750, M=1050, L=1200, XL=1600)
  plot.height <-   c(S=600, M=700,  L=900,  XL=1200)
  plot.height2 <-  c(S=150, M=200,  L=200,  XL=300) # for densities
  plot.fontsize <- c(S=12,  M=16,   L=18,   XL=24)
  plot.rows <-     c(S=2,   M=2,    L=3,    XL=3)
  observe({
    runjs(sprintf("$('#boxplot').css('width', %d).css('height', %d);",
                  plot.width[input$plot_size], plot.height[input$plot_size]))
    runjs(sprintf("$('#densities').css('width', %d).css('height', %d);",
                  plot.width[input$plot_size], plot.height2[input$plot_size]))
    runjs("$('#boxplot').trigger('resize'); $('#densities').trigger('resize');")
  })
  size.policy <- function(dims) {
    d <- abs(dims[1] - plot.width) + abs(dims[2] - plot.height)
    idx <- which.min(d) # find preset size closest to the dimensions reported by UI
    c(width=plot.width[idx], heigh=plot.height[idx])
  }
  
  output$boxplot <- renderCachedPlot({
    cat.variable <- switch(
      input$granularity,
      n12="textcat12", n20="textcat20", n32="textcat32")
    cat.selected <- switch(
      input$granularity,
      n12=input$show_textcat12, n20=input$show_textcat20, n32=input$show_textcat32)
    cat.colours <- switch(
      input$granularity,
      n12=rainbow12.vec, n20=rainbow20.vec, n32=rainbow32.vec)
    main.txt <- dim2label[input$dim]
    if (input$focus_variety != "all") main.txt <- sprintf("%s (ICE-%s)", main.txt, input$focus_variety)
    y.range <- if (input$y == "features") c(-1.5, 3) else c(-0.5, 1)
    plt <- ggbox.selected(Data()$M, Data()$Meta,
                          cats = cat.selected, 
                          variable = cat.variable,
                          colours = cat.colours,
                          weights = Data()$Dims[, input$dim],
                          what = input$y,
                          select = idx.weights,
                          feature.names = feature.names,
                          main = main.txt,
                          group.labels=!input$use_legend,
                          ylim = y.range,
                          nrow = plot.rows[input$plot_size], 
                          base_size = plot.fontsize[input$plot_size])
    if (!input$use_legend) plt <- plt + theme(axis.text.x=element_text(angle=52, hjust=1), legend.position="none") 
    plt
  }, 
  sizePolicy = size.policy,
  cacheKeyExpr = {
    list(variety = input$focus_variety,
         lda = input$lda,
         dim = input$dim,
         what = input$y,
         fine = input$granularity,
         size = input$plot_size,
         legend = input$use_legend,
         cats = switch(input$granularity, n12=input$show_textcat12, n20=input$show_textcat20, n32=input$show_textcat32))
  },
  cache = if (persistent.cache) "app" else "session")
  
  output$densities <- renderPlot({
    cat.vec <- switch(
      input$granularity,
      n12 = Data()$Meta$textcat12,
      n20 = Data()$Meta$textcat20,
      n32 = Data()$Meta$textcat32)
    cat.colours <- switch(
      input$granularity,
      n12=rainbow12.vec, n20=rainbow20.vec, n32=rainbow32.vec)
    cat.selected <- switch(
      input$granularity,
      n12 = input$show_textcat12, n20 = input$show_textcat20, n32 = input$show_textcat32)
    leg.cex <- switch(
      input$plot_size,
      S=0.7, M=0.9, L=1.0, XL=1.4)
    dim.vec <- Data()$Dims[, input$dim]
    idx <- if (length(cat.selected) > 0) cat.vec %in% cat.selected else NULL
    par(mar=c(2, 4, 0, 1))
    discriminant.plot(Data()$M, dim.vec, cat.vec, idx=idx,
                      col.vals=cat.colours, rug=TRUE,
                      legend.cex=leg.cex, legend.colsize=11,
                      y.max=2.2, xlim=c(-3, 5), xaxs="i", xlab="")
  },
  execOnResize = TRUE)
  
  output$downloadPDF <- downloadHandler(
    filename = function () {
      cat.list <- switch( 
        input$granularity,
        n12 = cat2label.12[input$show_textcat12],
        n20 = cat2label.20[input$show_textcat20],
        n32 = cat2label.32[input$show_textcat32])
      if (length(cat.list) > 0) {
        cat.info <- sprintf("_cat%d=%s", 
                            switch(input$granularity, n12=12, n20=20, n32=32),
                            paste(cat.list, collapse="-"))
      } else cat.info <- ""
      sprintf("ICE-boxplot_%s_%s_%s%s.pdf",
              input$focus_variety, 
              input$dim,
              input$y,
              cat.info)
    },
    content = function (filename) {
      cairo_pdf(filename, width=12, height=8)
      ## this just duplicates the code from renderCachedPlot() above
      cat.variable <- switch(
        input$granularity,
        n12="textcat12", n20="textcat20", n32="textcat32")
      cat.selected <- switch(
        input$granularity,
        n12=input$show_textcat12, n20=input$show_textcat20, n32=input$show_textcat32)
      cat.colours <- switch(
        input$granularity,
        n12=rainbow12.vec, n20=rainbow20.vec, n32=rainbow32.vec)
      main.txt <- dim2label[input$dim]
      if (input$focus_variety != "all") main.txt <- sprintf("%s (ICE-%s)", main.txt, input$focus_variety)
      y.range <- if (input$y == "features") c(-1.5, 3) else c(-0.5, 1)
      ggbox.selected(Data()$M, Data()$Meta,
                     cats = cat.selected, 
                     variable = cat.variable,
                     colours = cat.colours,
                     weights = Data()$Dims[, input$dim],
                     what = input$y,
                     select = idx.weights,
                     feature.names = feature.names,
                     main = main.txt,
                     ylim = y.range,
                     nrow = 2, base_size = 14) %>% print()
      dev.off()
    },
    contentType = "application/octet-stream"
  )
}

# Run the application 
shinyApp(ui = ui, server = server)
